
#include "data.h" //for debugging

//include this unit only when units needing it are included
#if defined(USE_I2C) 

#include <avr/io.h>
#include <avr/interrupt.h>
#include "TWI_Master.h"


#define NOSLAVE  /* cpu is master only */
#define MIN(a, b) (((a)<(b))?(a):(b))
#define EnterCritical()	sei()
#define ExitCritical()	cli()

#include "i2c_universal.h"

static char i2cerror=0, i2cstate=0, i2cindex=0;
static char *i2cdata = (char*)0, *i2cbuf = (char*)0;
static unsigned char TWI_state = TWI_NO_STATE;      // State byte. Default set to TWI_NO_STATE.


//The layout of the buffer s is: <nr of bytes incl. address byte(s)>, <slave address>, <1 or more sub-address or data bytes>...
char i2c_write(char s[]) 
{ if (s[0]<2)  //require at least address and one data cycle
  { return I2C_SHORT; //i2cerror;
  }
  if ((TWCR & _BV(TWIE)) || i2cstate) {
    return I2C_BUSY; //i2error;
  }
//  test(1);
//  EnterCritical();
  i2cerror = 0;
  i2cdata = s;
//  i2cdata[0]--;
//  i2cindex = 2;
  i2cstate = 1; //kick off whole process
  TWI_state         = TWI_NO_STATE ;
  TWCR = _BV(TWEN)|                             // TWI Interface enabled.
         _BV(TWIE)|_BV(TWINT)|                  // Enable TWI Interupt and clear the flag.
         _BV(TWSTA);                            // Initiate a START condition.
  //states to expect: 0x08
//  ExitCritical();
//  test(0);
  return 0; 
}

/* the layout of buffer s is: <number of bytes to write incl. addresses>, <slave address>, <0 or more sub_addresses>...
   when called s typically contains 2 to 4 bytes.
   the layout of buffer t is <number of bytes to read> <space for the bytes to read>...
   When reading has completed the first byte of s should be 0 and the other bytes contain the read data.
   The size of s should be 1 plus the number of bytes to read.
 */
   
char i2c_read(char s[], char t[]) 
{ if (s[0]<1 || t[0]<1)  //require at least an address (in write buffer) and one read cycle (to read buffer)
  {
    return I2C_SHORT; 
  }
  if ((TWCR & _BV(TWIE)) || i2cstate) {
    return I2C_BUSY;
  }
//  EnterCritical();
  i2cdata = s;
  i2cbuf  = t;
//  i2cindex = 2;//with reads, 1 contains the slave address
  i2cerror = 0; //SUCCESS
  if (s[0] == 1) // write only SLA => read only mode
    i2cstate = 3; //read only
  else
    i2cstate = 2; //write then read mode
  TWI_state         = TWI_NO_STATE ;
  TWCR = _BV(TWEN)|                             // TWI Interface enabled.
         _BV(TWIE)|_BV(TWINT)|                  // Enable TWI Interupt and clear the flag.
         _BV(TWSTA);                            // Initiate a START condition.
  //kick off whole process
//  ExitCritical();
  return 0;
}

/* the layout of buffer t is <number of bytes to read> <slave address> <space for the bytes to read>...
   When reading has completed the first byte of s should be 0 and the other bytes contain the read data.
   The size of t should be 1 plus the number of bytes to read (at least 1).
 */
   
char i2c_read_only(char t[]) 
{ if (t[0]<1)  //require at least an address and one read cycle (to read buffer)
  {
    return I2C_SHORT; 
  }
  if ((TWCR & _BV(TWIE)) || i2cstate) {
    return I2C_BUSY;
  }
  EnterCritical();
  i2cdata = 0;
  i2cbuf  = t;
//  i2cindex = 2;//with reads, 1 contains the slave address
  i2cerror = 0; //SUCCESS
  TWI_state         = TWI_NO_STATE ;
  TWCR = _BV(TWEN)|                             // TWI Interface enabled.
         _BV(TWIE)|_BV(TWINT)|                  // Enable TWI Interupt and clear the flag.
         _BV(TWSTA);                            // Initiate a START condition.
  //kick off whole process
  i2cstate = 4; //read only
  ExitCritical();
  return 0;
}

char device_busy(byte addr) //used to test if eeprom is busy writing, eeprom will not ack addr cycles when busy
{ char buf[2] = {1};
  if ((TWCR & _BV(TWIE)) || i2cstate) 
  { return I2C_BUSY;
  }
  buf[1] = addr;
  i2cdata = buf;
  EnterCritical();
  i2cerror = 0;
  TWI_state         = TWI_NO_STATE ;
  TWCR = _BV(TWEN)|                             // TWI Interface enabled.
         _BV(TWIE)|_BV(TWINT)|                  // Enable TWI Interupt and clear the flag.
         _BV(TWSTA);                            // Initiate a START condition.
  i2cstate = 1; //kick off whole process
  ExitCritical();
  while (i2cstate); //busy wait !!, about 1 byte time (approx. 0.1 ms @ 100kHz)
  return i2cerror; 
}
  
bool i2c_busy()
{ return i2cstate;
}

char i2clasterror()
{ //return i2cstate?I2C_BUSY:i2cerror;
  return i2cerror;
}

// only master read/write fully implemented, other modes just framework

ISR(TWI_vect)
{
  switch (TWSR & 0xF8)
  {
    case TWI_START:             // START has been transmitted 
	      if (i2cstate == 3) //read mode
		  { i2cindex = 1;
		    TWDR = i2cdata[1] | 1; //SLA+R
		  } 
		  else if (i2cstate == 4) //read_only mode
		  { i2cindex = 1;
		    TWDR = i2cbuf[1] | 1; //SLA+R
		  } 		   
		  else // write mode or write then read mode
		  { i2cindex = 2;
		    TWDR = i2cdata[1] & ~1; //SLA+W
          }
          TWCR = _BV(TWEN)|                                 // TWI Interface enabled
                 _BV(TWIE)|_BV(TWINT);                      // Enable TWI Interupt and clear the flag to send byte
 		  break; //next states TWI_MTX_ADR_ACK/NACK
    case TWI_REP_START:         // Repeated START has been transmitted
          i2cindex = 1;         // Set buffer pointer to the TWI Address location
		  TWDR = i2cdata[1] | 1; //SLA+R
          TWCR = _BV(TWEN)|                                 // TWI Interface enabled
                 _BV(TWIE)|_BV(TWINT);                      // Enable TWI Interupt and clear the flag to send byte
 		  break; //next states TWI_MRX_ADR_ACK/NACK
    case TWI_MTX_ADR_ACK:       // SLA+W has been tramsmitted and ACK received
    case TWI_MTX_DATA_ACK:      // Data byte has been tramsmitted and ACK received
      if (i2cindex <= i2cdata[0])
      {
        TWDR = i2cdata[i2cindex++];
        TWCR = _BV(TWEN)|                                 // TWI Interface enabled
               _BV(TWIE)|_BV(TWINT);                      // Enable TWI Interupt and clear the flag to send byte
                                                //  
      }else                   
      { if (i2cstate == 2) //write then read mode, switch to read
	    { //generate repeated start condition
          TWCR = _BV(TWEN)|                             // TWI Interface enabled.
                 _BV(TWIE)|_BV(TWINT)|                  // Enable TWI Interupt and clear the flag.
                 _BV(TWSTA);                            // Initiate a START condition.
		}
		else  // Send STOP after last byte
		{ i2cstate = 0;                 // Set status bits to completed successfully.
		  i2cdata[0] = 0; //indicate that buffer is free 
          TWCR = _BV(TWEN)|                                 // TWI Interface enabled
                 _BV(TWINT)|                      // Disable TWI Interrupt and clear the flag
                 _BV(TWSTO);           // Initiate a STOP condition.
		}
      }
      break;
    case TWI_MRX_DATA_ACK:      // Data byte has been received and ACK tramsmitted
      i2cbuf[i2cindex++] = TWDR;
    case TWI_MRX_ADR_ACK:       // SLA+R has been tramsmitted and ACK received
      if (i2cindex < i2cbuf[0] )                  // Detect the last byte to NACK it.
      {
        TWCR = _BV(TWEN)|                                 // TWI Interface enabled
               _BV(TWIE)|_BV(TWINT)|                      // Enable TWI Interupt and clear the flag to read next byte
               _BV(TWEA);           // Send ACK after reception
      }else                    // Send NACK after next reception
      {
        TWCR = _BV(TWEN)|                                 // TWI Interface enabled
               _BV(TWIE)|_BV(TWINT);                      // Enable TWI Interupt and clear the flag to read next byte
                                                          // Send NACK after reception
      }    
      break; 
    case TWI_MRX_DATA_NACK:     // Data byte has been received and NACK tramsmitted
      i2cbuf[i2cindex] = TWDR;
      i2cstate = 0;                // Set status bits to completed successfully. 
      i2cdata[0] = 0;  //indicate that buffers are free
      //i2cbuf[0] = 0;
      TWCR = _BV(TWEN)|                                 // TWI Interface enabled
             _BV(TWINT)|                      // Disable TWI Interrupt and clear the flag
             _BV(TWSTO);           // Initiate a STOP condition.
      break;      
    case TWI_ARB_LOST:          // Arbitration lost
      TWCR = (1<<TWEN)|                                 // TWI Interface enabled
             (1<<TWIE)|(1<<TWINT)|                      // Enable TWI Interupt and clear the flag
             (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|           // Initiate a (RE)START condition.
             (0<<TWWC);                                 //
      break;
    case TWI_MTX_ADR_NACK:      // SLA+W has been tramsmitted and NACK received
    case TWI_MRX_ADR_NACK:      // SLA+R has been tramsmitted and NACK received    
	   i2cerror = I2C_NODEV;
    case TWI_MTX_DATA_NACK:     // Data byte has been tramsmitted and NACK received
	   if (!i2cerror)
	     i2cerror = I2C_ACK;
//    case TWI_NO_STATE              // No relevant state information available; TWINT = 0
    case TWI_BUS_ERROR:         // Bus error due to an illegal START or STOP condition
    default:
      TWI_state = TWSR;                                 // Store TWSR and automatically sets clears noErrors bit.
                                                        // Reset TWI Interface
      TWCR = _BV(TWEN)|_BV(TWINT)|_BV(TWSTO);           // Enable TWI-interface and release TWI pins
                                                        // Disable Interupt
                                                        // No Signal requests
      if (i2cstate > 1) //read mode
	    i2cbuf[0] = 0;
      i2cdata[0] = 0;
      i2cstate = 0;
   }
}

//#define SCL_FREQ	360000
#define SCL_FREQ	100000
void IIC1_Init(void)
{ //prescaler = 1
#if F_CPU == 18432000L
  TWBR = 84; //100.17kHz
#elif F_CPU == 8000000L
  TWBR = 32;
#elif F_CPU == 12000000L
  TWBR = 52;
#elif F_CPU == 9830400L
  TWBR = 41; //100.3kHz
#elif F_CPU == 14745600L
  TWBR = (F_CPU/SCL_FREQ - 14)/2; //rounded up, hence Fscl is rounded down
#else
  TWBR = 92; 
#error "unknown cpu freq"
#endif
//  TWCR = _BV(TWIE) | _BV(TWSTO) | _BV(TWINT);
  TWCR = _BV(TWEN);
  TWSR = 0;                 
}

#endif //inclusion of this module

/* END i2c */
